/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */

/* UNIX INCLUDES */
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdio.h>

/* PVFS INCLUDES */
#include <build_job.h>
#include <build_list_job.h>
#include <desc.h>
#include <req.h>
#include <jlist.h>
#include <alist.h>
#include <minmax.h>
#include <sockset.h>
#include <debug.h>
#include <sockio.h>
#include <dfd_set.h>

/* GLOBALS */
extern jlist_p active_p;
extern sockset socks;

#define PCOUNT ((int64_t)(f_p->fd.meta.p_stat.pcount))

/* validate no negative list values */
int list_check(int     mem_list_count,
					int     mem_lengths[],
					int     file_list_count,
					int64_t file_offsets[],
					int32_t file_lengths[])
{
	int i;

	if (mem_list_count <= 0) {
		errno = EINVAL;
		return -1;
	}
	if (file_list_count <= 0) {
		errno = EINVAL;
		return -1;
	}
	for (i = 0; i < mem_list_count; i++) {
      if (mem_lengths[i] < 0) {
			errno = EINVAL;
			return -1;
		}
	}
	for (i = 0; i < file_list_count; i++) {
      if (file_offsets[i] < 0) {
			errno = EINVAL;
			return -1;
		}
      if (file_lengths[i] < 0) {
			errno = EINVAL;
			return -1;
		}     
	}
	return 0;
}

/* check to make sure that the total memory size and the total file size 
 * are equal 
 */
int check_list_size(int mem_list_count, 
						  int mem_lengths[],
						  int file_list_count, 
						  int32_t file_lengths[])
{
	int i, mem_temp, file_temp;
  
	mem_temp = 0;
	file_temp = 0;

	/* add up the total memory size */
	for (i = 0; i < mem_list_count; i++)
	{
      mem_temp += mem_lengths[i];
	}

	/* add up the total file size */
	for (i = 0; i < file_list_count; i++)
	{
      file_temp += file_lengths[i];
	}

	if (mem_temp != file_temp)
	{
		errno = EINVAL;
		return -1;
	}

	/* return the total size */
	return mem_temp;
}

/* BUILD_LIST_JOBS() - construct jobs to fulfill an I/O request
 * PARAMETERS:
 *    f_p   - pointer to file descriptor
 *    type  - type of jlist request
 *
 * Returns -1 on error, 0 on success?
 */
int build_list_jobs(fdesc_p f_p, 
						  int     mem_list_count, 
						  char   *mem_offsets[],
						  int     mem_lengths[],
						  int     file_list_count,
						  int64_t file_offsets[],
						  int32_t file_lengths[],
						  int     type)
{
	ireq    req;
	ireq_p  r_p = &req; /* nice way to use the same defines throughout */
	ainfo_p a_p;
	jinfo_p j_p;
	int i, atype, sock, myerr;

	long long *file_offsets_p; /* pointers for arrays*/
	int *file_lengths_p;

	int total_size, tmp_size; /* variables for add_accesses*/
	int mem_num, file_num, mem_used, file_used, mem_remain, file_remain;
	char *tmp_mem_ptr;
	int tmp_file_off;

	/* build the requests */
	if (!active_p) active_p = jlist_new();
	LOG3("build_list_jobs: a = %lx, s = %Ld, t = %d\n", buf_p, size, type);

	build_list_req(f_p, &req, file_list_count, type);
	LOG("  build_list_jobs: done building request\n");

	/* sooner or later this will matter... */
	switch (type) {
	case J_READ:
		atype = A_READ; 
		break;
	case J_WRITE:
	default:
		atype = A_WRITE;
		break;
	}

	/* find the total length of all the accesses and check them */
	total_size = check_list_size(mem_list_count, mem_lengths, 
										  file_list_count, file_lengths);
	if (total_size < 0) {
		/* errno set by check_list_size */
		return -1;
	}

	/* add the accesses */ /* initialize the ptrs */
	tmp_mem_ptr = (char *) mem_offsets[0];
	tmp_file_off = file_offsets[0];
	
	tmp_size = 0;
	mem_used = 0;
	file_used = 0;
	mem_num = 0;
	file_num = 0;

	while (tmp_size < total_size)
	{
		/* remaining number of bytes for current contiguous regions */
		mem_remain = mem_lengths[mem_num] - mem_used;
		file_remain = file_lengths[file_num] - file_used;

		/* case: mem_length = file_length */
		if (mem_remain == file_remain)
		{
			add_accesses(f_p, tmp_file_off, file_remain, tmp_mem_ptr, atype);
			tmp_size += file_remain;
			mem_num++;
			file_num++;
			mem_used = 0;
			file_used = 0;
			tmp_mem_ptr = (char *) mem_offsets[mem_num];
			tmp_file_off = file_offsets[file_num];
		}
		/* case: mem_length < file_length */
		else if (mem_remain < file_remain)
		{
			add_accesses(f_p, tmp_file_off, mem_remain, tmp_mem_ptr, atype);
			tmp_size += mem_remain;
			file_used += mem_remain;
			tmp_file_off += mem_remain;
			mem_num++;
			mem_used = 0;	
			tmp_mem_ptr = (char *) mem_offsets[mem_num];
		}
		/* case: mem_length > file_length */
		else
		{
			add_accesses(f_p, tmp_file_off, file_remain, tmp_mem_ptr, atype);
			tmp_size += file_remain;
			mem_used += file_remain;
			tmp_mem_ptr += file_remain;
			file_num++;
			file_used = 0;
			tmp_file_off = file_offsets[file_num];
		}
	}
	
	LOG("build_list_jobs: done calling add_accesses\n");
	  
	initset(&socks); /* clear any other jobs out */
	for (i = 0; i < PCOUNT; i++) {
		sock = f_p->fd.iod[i].sock;
		if (sock >= 0 && (j_p = j_search(active_p, sock)) != NULL) {
	      /* add reception of ack */
	      if ((a_p = (ainfo_p) malloc(sizeof(ainfo))) == NULL) {
				PERROR("build_list_jobs: malloc (ainfo1)");
				return(-1);
	      }
	      a_p->type = A_ACK;
	      a_p->sock = sock;
	      a_p->u.ack.size = sizeof(iack);
	      a_p->u.ack.ack_p = a_p->u.ack.cur_p = (char *)&(f_p->fd.iod[i].ack);
	      a_p->u.ack.iod_nr = i;
	      
	      switch(type) {
	      case J_READ:
				if (a_add_start(j_p->al_p, a_p) < 0) {
					myerr = errno;
					ERR("build_list_jobs: a_add_start failed (ack)\n");
					errno = myerr;
					return(-1);
				}
				break;
	      case J_WRITE:
	      default:
				if (a_add_end(j_p->al_p, a_p) < 0) {
					myerr = errno;
					ERR("build_list_jobs: a_add_end failed\n");
					errno = myerr;
					return(-1);
				}
				break;
	      }
	      LOG1("build_list_jobs: added ack recv for iod %d\n", i);
			r_p = (ireq_p) malloc(sizeof(ireq) + 
										 file_list_count*sizeof(int64_t) +
										 file_list_count*sizeof(int32_t));
			if (r_p == NULL) {
				myerr = errno;
				PERROR("build_list_jobs: malloc (ireq)");
				errno = myerr;
				return(-1);
			}
			file_offsets_p = (int64_t *) ((char *)r_p + sizeof(ireq));
			file_lengths_p = (int32_t *) ((char *)file_offsets_p + 
													file_list_count*sizeof(int64_t));

			*r_p = req;
			memcpy(file_offsets_p, file_offsets, file_list_count*sizeof(int64_t));
			memcpy(file_lengths_p, file_lengths, file_list_count*sizeof(int32_t));
		
			if ((a_p = (ainfo_p) malloc(sizeof(ainfo))) == NULL) {
				myerr = errno;
				PERROR("build_list_jobs: malloc (ainfo2)");
				errno = myerr;
				return(-1);
			}
			a_p->type = A_REQ;
			a_p->sock = sock;
			/* includes ireq and 2 arrays */
			a_p->u.req.size = sizeof(ireq) + file_list_count*sizeof(int64_t) 
				+ file_list_count*sizeof(int32_t);
			a_p->u.req.cur_p = a_p->u.req.req_p = (char *) r_p;
			if (a_add_start(j_p->al_p, a_p) < 0) {
				myerr = errno;
				ERR("build_list_jobs: a_add_start failed (req)\n");
				errno = myerr;
				return(-1);
			}
			addsockwr(sock, &socks);
			LOG1("build_list_jobs: added req send for iod %d\n", i);
		} /* end of requesting process code */
		else /* get ready to read or write */ {
			switch(type) {
			case J_READ:
				addsockrd(sock, &socks);
				break;
			case J_WRITE:
				addsockwr(sock, &socks);
				break;
			default:
				ERR("build_list_jobs: bad job type\n");
			}
		}
		/* else there's no job for this socket */
	}
	LOG("build_list_jobs: done.\n");
	return 0;
}

#undef PCOUNT

/* BUILD_LIST_REQ() - Builds a request for an application 
  *
 * Note: num parameter was only used in group I/O requests and corresponds
 * to the number of the process (0..n)
 *
 * NOTE: NUM IS NO LONGER USED.
 *
 * Also: type is in terms of a jlist type; it must be converted to an 
 * IOD_RW type
 */
int build_list_req(fdesc_p f_p, 
						 ireq_p  r_p,
						 int     file_list_count,
						 int     type)
{
	/* set up request */
	r_p->majik_nr = IOD_MAJIK_NR;
	r_p->type     = IOD_LIST;
	/* dsize includes a list of offsets and one of lengths */
	r_p->dsize    = file_list_count*sizeof(int32_t) + 
		file_list_count*sizeof(int64_t);
	r_p->req.list.f_ino = f_p->fd.meta.u_stat.st_ino;
	r_p->req.list.cap   = f_p->fd.cap;
	r_p->req.list.file_list_count = file_list_count; 

	switch (type) {
	case J_READ:
		r_p->req.list.rw_type = IOD_RW_READ;
		break;
	case J_WRITE:
		r_p->req.list.rw_type = IOD_RW_WRITE;
		break;
	default:
		r_p->req.list.rw_type = IOD_RW_WRITE;
		break;
	}

#ifdef DEBUG
	if (debug_on) fprintf(stderr,
								 "file_list_count = %d; type = %d\n", 
								 file_list_count, type);
#endif
	return 0;
} /* end of BUILD_LIST_REQ() */

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 * End:
 */
